"""
SPEC Data File Writer Callback
+++++++++++++++++++++++++++++++++++++++

EXAMPLE:

Execution of this plan (with ``RE(myPlan())``)::

    def myPlan():
        yield from bps.open_run()
        spec_comment("this is a start document comment", "start")
        spec_comment("this is a descriptor document comment", "descriptor")
        yield bps.Msg('checkpoint')
        yield from bps.trigger_and_read([scaler])
        spec_comment("this is an event document comment after the first read")
        yield from bps.sleep(2)
        yield bps.Msg('checkpoint')
        yield from bps.trigger_and_read([scaler])
        spec_comment("this is an event document comment after the second read")
        spec_comment("this is a stop document comment", "stop")
        yield from bps.close_run()

results in this SPEC file output::

    #S 1145  myPlan()
    #D Mon Jan 28 12:48:09 2019
    #C Mon Jan 28 12:48:09 2019.  plan_type = generator
    #C Mon Jan 28 12:48:09 2019.  uid = ef98648a-8e3a-4e7e-ac99-3290c9b5fca7
    #C Mon Jan 28 12:48:09 2019.  this is a start document comment
    #C Mon Jan 28 12:48:09 2019.  this is a descriptor document comment
    #MD APSTOOLS_VERSION = 2019.0103.0+5.g0f4e8b2
    #MD BLUESKY_VERSION = 1.4.1
    #MD OPHYD_VERSION = 1.3.0
    #MD SESSION_START = 2019-01-28 12:19:25.446836
    #MD beamline_id = developer
    #MD ipython_session_start = 2018-02-14 12:54:06.447450
    #MD login_id = mintadmin@mint-vm
    #MD pid = 21784
    #MD proposal_id = None
    #N 2
    #L Epoch_float  scaler_time  Epoch
    1.4297869205474854 1.1 1
    4.596935987472534 1.1 5
    #C Mon Jan 28 12:48:11 2019.  this is an event document comment after the first read
    #C Mon Jan 28 12:48:14 2019.  this is an event document comment after the second read
    #C Mon Jan 28 12:48:14 2019.  this is a stop document comment
    #C Mon Jan 28 12:48:14 2019.  num_events_primary = 2
    #C Mon Jan 28 12:48:14 2019.  exit_status = success


.. autosummary::

   ~SpecWriterCallback
   ~spec_comment
"""

from collections import OrderedDict
import datetime
import getpass
import logging
import numpy as np
import os
import socket
import time


SPEC_TIME_FORMAT = "%a %b %d %H:%M:%S %Y"
SCAN_ID_RESET_VALUE = 0


def _rebuild_scan_command(doc):
    """
    reconstruct the scan command for SPEC data file #S line

    PARAMETERS

    doc
        *object* :
        instance of a bluesky ``start`` document

    RETURNS

    *str* :
        "scan_number reconstructed_scan_command"
    """

    def get_name(src):
        """
        get name field from object representation

        given: EpicsMotor(prefix='xxx:m1', name='m1', settle_time=0.0,
                    timeout=None, read_attrs=['user_readback', 'user_setpoint'],
                    configuration_attrs=['motor_egu', 'velocity', 'acceleration',
                    'user_offset', 'user_offset_dir'])
        return: "m1"
        """
        s = str(src)
        p = s.find("(")
        if p > 0:  # only if an open parenthesis is found
            # fmt: offf
            # fmt: on
            parts = s[p + 1:].rstrip(")").split(",")
            for item in parts:
                # should be key=value pairs
                item = item.strip()
                p = item.find("=")
                if item[:p] == "name":
                    # fmt: off
                    # fmt: on
                    s = item[p + 1:]  # get the name value
                    break
        return s

    def struct_to_str(struct):
        """Convert given structure into string representation."""
        if isinstance(struct, list):
            return (
                "[" + ", ".join([struct_to_str(v) for v in struct]) + "]"
            )
        elif isinstance(struct, dict):
            return (
                "{"
                + ", ".join(
                    [f"{k}={struct_to_str(v)}" for k, v in struct.items()]
                )
                + "}"
            )
        elif isinstance(struct, np.ndarray):
            return struct_to_str(list(struct))
        elif isinstance(struct, str):
            # escape embedded characters, such as single quote
            return f"'{struct}'".encode("unicode_escape").decode("utf8")
        else:
            return str(struct)

    s = []
    if "plan_args" in doc:
        for _k, _v in doc["plan_args"].items():
            if _k == "detectors":
                _v = doc[_k]
            elif _k.startswith("motor"):
                _v = doc["motors"]
            elif _k == "args":
                _v = "[" + ", ".join(map(get_name, _v)) + "]"
            s.append(f"{_k}={struct_to_str(_v)}")

    cmd = "{}({})".format(doc.get("plan_name", ""), ", ".join(s))
    scan_id = doc.get("scan_id") or 1
    return f"{scan_id}  {cmd}"


# TODO: consider refactor to use FileWriterCallbackBase()
class SpecWriterCallback(object):
    """
    Collect data from Bluesky RunEngine documents to write as SPEC data.

    .. index:: Bluesky Callback; SpecWriterCallback

    This gathers data from all documents in a scan and appends scan to the file
    when the ``stop`` document is received.  One or more scans can be written to
    the same file.  The file format is text.

    .. note:: ``SpecWriterCallback()`` does **not** inherit
       from ``FileWriterCallbackBase()``.

    PARAMETERS

    filename
        *string* :
        (optional)
        Local, relative or absolute name of SPEC data file to be used.
        If ``filename=None``, defaults to format of ``YYYmmdd-HHMMSS.dat``
        derived from the current system time.

    auto_write
        *boolean* :
        (optional)
        If ``True`` (default), ``write_scan()`` is called when *stop* document
        is received.
        If ``False``, the caller is responsible for calling ``write_scan()``
        before the next ``start`` document is received.

    RE
        *object* :
        Instance of ``bluesky.RunEngine`` or ``None``.

    reset_scan_id
        *boolean* :
        (optional)
        If True, and filename exists, then sets ``RE.md.scan_id`` to
        highest scan number in existing SPEC data file.
        default: False

    User Interface methods

    .. autosummary::

       ~receiver
       ~newfile
       ~usefile
       ~make_default_filename
       ~clear
       ~prepare_scan_contents
       ~write_scan

    Internal methods

    .. autosummary::

       ~write_header
       ~start
       ~descriptor
       ~event
       ~bulk_events
       ~datum
       ~resource
       ~stop
    """

    def __init__(
        self, filename=None, auto_write=True, RE=None, reset_scan_id=False
    ):
        self.clear()
        self.buffered_comments = self._empty_comments_dict()
        self.spec_filename = filename
        self.auto_write = auto_write
        self.uid_short_length = 8
        self.write_file_header = False
        self.spec_epoch = None  # for both #E & #D line in header, also offset for all scans
        self.spec_host = None
        self.spec_user = None
        self._datetime = None  # most recent document time
        self._streams = {}  # descriptor documents, keyed by uid
        self.RE = RE

        if reset_scan_id is True:
            reset_scan_id = SCAN_ID_RESET_VALUE
        self.reset_scan_id = reset_scan_id

        if filename is None or not os.path.exists(filename):
            self.newfile(filename)
        else:
            max_scan_id = self.usefile(filename)
            if RE is not None and reset_scan_id is not False:
                RE.md["scan_id"] = max_scan_id

    def clear(self):
        """reset all scan data defaults"""
        self.uid = None
        self.scan_epoch = None  # absolute epoch to report in scan #D line
        self.time = None  # full time from document
        self.comments = self._empty_comments_dict()
        self.data = OrderedDict()  # data in the scan
        self.detectors = OrderedDict()  # names of detectors in the scan
        self.hints = OrderedDict()  # why?
        self.metadata = OrderedDict()  # #MD lines in header
        self.motors = OrderedDict()  # names of motors in the scan
        self.positioners = OrderedDict()  # names in #O, values in #P
        self.num_primary_data = 0
        #
        # note: for one scan, #O & #P information is not provided
        # unless collecting baseline data
        # wait for case with baseline data that needs #O/#P lines
        #
        self.columns = OrderedDict()  # #L in scan
        self.scan_command = None  # #S line
        self.scanning = False

    def _empty_comments_dict(self):
        return dict(
            start=[],
            event=[],
            descriptor=[],
            resource=[],
            datum=[],
            stop=[],
        )

    def _cmt(self, key, text):
        """enter a comment"""
        dt = self._datetime or datetime.datetime.now()
        ts = datetime.datetime.strftime(dt, SPEC_TIME_FORMAT)
        if self.scanning:
            dest = self.comments
        else:
            dest = self.buffered_comments
        dest[key].append(f"{ts}.  {text}")

    def receiver(self, key, document):
        """
        Bluesky callback: receive all documents for handling

        .. index:: Bluesky Callback; SpecWriterCallback.receiver
        """
        xref = dict(
            start=self.start,
            descriptor=self.descriptor,
            event=self.event,
            bulk_events=self.bulk_events,
            datum=self.datum,
            resource=self.resource,
            stop=self.stop,
        )
        logger = logging.getLogger(__name__)
        if key in xref:
            token = document.get("uid") or document.get("datum_id")
            logger.debug("%s document, uid=%s", key, str(token))  # lgtm [py/clear-text-logging-sensitive-data]
            ts = document.get("time")
            if ts is None:
                ts = datetime.datetime.now()
            else:
                ts = datetime.datetime.fromtimestamp(document["time"])
            self._datetime = ts
            xref[key](document)
        else:
            msg = f"custom_callback encountered: {key} : {document}"
            # raise ValueError(msg)
            logger.warning(msg)

    def start(self, doc):
        """handle *start* documents"""

        known_properties = """
            uid time project sample scan_id group owner
            hints
            plan_type plan_name plan_args
        """.split()

        self.clear()
        self.scanning = True
        self.uid = doc["uid"]

        self._cmt("start", f"uid = {self.uid}")
        self.metadata["uid"] = f"{self.uid}"
        for d, cl in self.buffered_comments.items():
            # bring in any comments collected when not scanning
            self.comments[d] += cl
        self.buffered_comments = self._empty_comments_dict()

        self.time = doc["time"]
        self.scan_epoch = int(self.time)
        self.scan_id = doc["scan_id"] or 0
        # Which reference? fixed counting time or fixed monitor count?
        # Can this be omitted?
        self.T_or_M = None  # for now
        # self.T_or_M = "T"           # TODO: how to get this from the document stream?
        # self.T_or_M_value = 1
        # self._cmt("start", "!!! #T line not correct yet !!!")

        # metadata
        for key in sorted(doc.keys()):
            if key not in known_properties:
                self.metadata[key] = doc[key]

        self.start_hints = doc.get("hints", {})

        # various dicts
        for item in "detectors hints motors".split():
            if item in doc:
                obj = self.__getattribute__(item)
                for key in doc.get(item):
                    obj[key] = None

        cmt = "plan_type = " + doc["plan_type"]
        ts = datetime.datetime.strftime(self._datetime, SPEC_TIME_FORMAT)
        self.comments["start"].insert(0, f"{ts}.  {cmt}")
        self.scan_command = _rebuild_scan_command(doc)

    def descriptor(self, doc):
        """
        handle *descriptor* documents

        prepare for primary scan data, ignore any other data stream
        """
        if doc["uid"] in self._streams:
            fmt = "duplicate descriptor UID {} found"
            raise KeyError(fmt.format(doc["uid"]))

        # log descriptor documents by uid
        # referenced by event and bulk_events documents
        self._streams[doc["uid"]] = doc

        if doc["name"] != "primary":
            return

        keyset = list(doc["data_keys"].keys())
        doc_hints_names = []
        for k, d in doc["hints"].items():
            doc_hints_names.append(k)
            doc_hints_names += d["fields"]

        # independent variable(s) first
        # assumes start["motors"] was defined
        first_keys = [k for k in self.motors if k in keyset]
        # TODO: if len(first_keys) == 0: look at self.start_hints

        # dependent variable(s) last
        # assumes start["detectors"] was defined
        last_keys = [d for d in self.detectors if d in doc_hints_names]
        # TODO: if len(last_keys) == 0: look at doc["hints"]

        # get remaining keys from keyset, they go in the middle
        middle_keys = [
            k for k in keyset if k not in first_keys + last_keys
        ]
        epoch_keys = "Epoch_float Epoch".split()

        self.data.update(
            {
                k: []
                for k in first_keys + epoch_keys + middle_keys + last_keys
            }
        )

    def event(self, doc):
        """
        handle *event* documents
        """
        stream_doc = self._streams.get(doc["descriptor"])
        if stream_doc is None:
            fmt = "descriptor UID {} not found"
            raise KeyError(fmt.format(doc["descriptor"]))
        if stream_doc["name"] == "primary":
            for k in doc["data"].keys():
                if k not in self.data.keys():
                    msg = f"unexpected failure here, key {k} not found"
                    raise KeyError(msg)
            for k in self.data.keys():
                if k == "Epoch":
                    v = int(doc["time"] - self.time + 0.5)
                elif k == "Epoch_float":
                    v = doc["time"] - self.time
                else:
                    # like SPEC, default to 0 if not found by name
                    v = doc["data"].get(k, 0)
                self.data[k].append(v)
            self.num_primary_data += 1

    def bulk_events(self, doc):
        """handle *bulk_events* documents"""
        pass

    def datum(self, doc):
        """handle *datum* documents"""
        self._cmt("datum", "datum " + str(doc))

    def resource(self, doc):
        """handle *resource* documents"""
        self._cmt("resource", "resource " + str(doc))

    def stop(self, doc):
        """handle *stop* documents"""
        if "num_events" in doc:
            for k, v in doc["num_events"].items():
                self._cmt("stop", f"num_events_{k} = {v}")
        if "exit_status" in doc:
            self._cmt("stop", "exit_status = " + doc["exit_status"])
        else:
            self._cmt("stop", "exit_status = not available")

        if self.auto_write:
            self.write_scan()

        self.scanning = False

    def prepare_scan_contents(self):
        """
        format the scan for a SPEC data file

        :returns: [str] a list of lines to append to the data file
        """
        dt = datetime.datetime.fromtimestamp(self.scan_epoch)
        lines = []
        lines.append("")
        lines.append("#S " + self.scan_command)
        lines.append(
            "#D " + datetime.datetime.strftime(dt, SPEC_TIME_FORMAT)
        )
        if self.T_or_M is not None:
            lines.append(f"#{self.T_or_M} {self.T_or_M_value}")

        for v in self.comments["start"]:
            # C Wed Feb 03 16:51:38 2016.  do ./usaxs.mac.
            lines.append(
                "#C " + v
            )  # TODO: add time/date stamp as SPEC does
        for v in self.comments["descriptor"]:
            lines.append("#C " + v)

        for k, v in self.metadata.items():
            # "#MD" is our ad hoc SPEC data tag
            lines.append(f"#MD {k} = {v}")

        lines.append("#P0 ")

        lines.append("#N " + str(len(self.data.keys())))
        if len(self.data.keys()) > 0:
            lines.append("#L " + "  ".join(self.data.keys()))
            for i in range(self.num_primary_data):
                str_data = OrderedDict()
                s = []
                for k in self.data.keys():
                    datum = self.data[k][i]
                    if isinstance(datum, str):
                        # SPEC scan data is expected to be numbers
                        # this is text, substitute the row number
                        # and report after this line in a #U line
                        str_data[k] = datum
                        datum = i
                    s.append(str(datum))
                lines.append(" ".join(s))
                for k in str_data.keys():
                    # report the text data
                    lines.append(f"#U {i} {k} {str_data[k]}")
        else:
            lines.append("#C no data column labels identified")

        for v in self.comments["event"]:
            lines.append("#C " + v)

        for v in self.comments["resource"]:
            lines.append("#C " + v)

        for v in self.comments["datum"]:
            lines.append("#C " + v)

        for v in self.comments["stop"]:
            lines.append("#C " + v)

        return lines

    def _write_lines_(self, lines, mode="a"):
        """write (more) lines to the file"""
        with open(self.spec_filename, mode) as f:
            f.write("\n".join(lines))

    def write_header(self):
        """write the header section of a SPEC data file"""
        dt = datetime.datetime.fromtimestamp(self.spec_epoch)
        lines = []
        lines.append(f"#F {self.spec_filename}")
        lines.append(f"#E {self.spec_epoch}")
        lines.append(
            f"#D {datetime.datetime.strftime(dt, SPEC_TIME_FORMAT)}"
        )
        lines.append(
            f"#C Bluesky  user = {self.spec_user}  host = {self.spec_host}"
        )
        lines.append("#O0 ")
        lines.append("#o0 ")
        lines.append("")

        if os.path.exists(self.spec_filename):
            lines.insert(0, "")
        self._write_lines_(lines, mode="a+")
        self.write_file_header = False

    def write_scan(self):
        """
        write the most recent (completed) scan to the file

        * creates file if not existing
        * writes header if needed
        * appends scan data

        note:  does nothing if there are no lines to be written
        """
        if os.path.exists(self.spec_filename):
            with open(self.spec_filename) as f:
                buf = f.read()
                if buf.find(self.uid) >= 0:
                    # raise exception if uid is already in the file!
                    msg = f"{self.spec_filename} already contains uid={self.uid}"
                    raise ValueError(msg)
        logger = logging.getLogger(__name__)
        lines = self.prepare_scan_contents()
        lines.append("")
        if lines is not None:
            if self.write_file_header:
                self.write_header()
                logger.info(
                    "wrote header to SPEC file: %s", self.spec_filename
                )
            self._write_lines_(lines, mode="a")
            logger.info(
                "wrote scan %d to SPEC file: %s",
                self.scan_id,
                self.spec_filename,
            )

    def make_default_filename(self):
        """generate a file name to be used as default"""
        now = datetime.datetime.now()
        return datetime.datetime.strftime(now, "%Y%m%d-%H%M%S") + ".dat"

    def newfile(self, filename=None, scan_id=None, RE=None):
        """
        prepare to use a new SPEC data file

        but don't create it until we have data
        """
        self.clear()
        filename = filename or self.make_default_filename()
        if os.path.exists(filename):
            from spec2nexus.spec import SpecDataFile

            sdf = SpecDataFile(filename)
            scan_list = sdf.getScanNumbers()
            l = len(scan_list)
            m = max(map(float, scan_list))
            highest = int(max(l, m) + 0.9999)  # solves issue #128
            scan_id = max(scan_id or 0, highest)
        self.spec_filename = filename
        self.spec_epoch = int(time.time())  # ! no roundup here!!!
        self.spec_host = socket.gethostname() or "localhost"
        self.spec_user = getpass.getuser() or "BlueskyUser"
        self.write_file_header = True  # don't write the file yet

        # backwards-compatibility
        if isinstance(scan_id, bool):
            # True means reset the scan ID to default
            # False means do not modify it
            scan_id = {True: SCAN_ID_RESET_VALUE, False: None}[scan_id]
        if scan_id is not None and RE is not None:
            # RE is an instance of bluesky.run_engine.RunEngine
            # (or duck type for testing)
            RE.md["scan_id"] = scan_id
            self.scan_id = scan_id
        return self.spec_filename

    def usefile(self, filename):
        """read from existing SPEC data file"""
        if not os.path.exists(self.spec_filename):
            raise IOError(f"file {filename} does not exist")
        scan_id = None
        with open(filename, "r") as f:
            key = "#F"
            line = f.readline().strip()
            if not line.startswith(key + " "):
                raise ValueError(f"first line does not start with {key}")

            key = "#E"
            line = f.readline().strip()
            if not line.startswith(key + " "):
                raise ValueError(f"first line does not start with {key}")
            epoch = int(line.split()[-1])

            key = "#D"
            line = f.readline().strip()
            if not line.startswith(key + " "):
                raise ValueError("first line does not start with " + key)
            # ignore content, it is derived from #E line

            key = "#C"
            line = f.readline().strip()
            if not line.startswith(key + " "):
                raise ValueError("first line does not start with " + key)
            p = line.split()
            username = "BlueskyUser"
            if len(p) > 4 and p[2] == "user":
                username = p[4]

            # find the highest scan number used
            key = "#S"
            scan_ids = []
            for line in f.readlines():
                if line.startswith(key + " ") and len(line.split()) > 1:
                    scan_id = int(line.split()[1])
                    scan_ids.append(scan_id)
            scan_id = max(scan_ids)

        self.spec_filename = filename
        self.spec_epoch = epoch
        self.spec_user = username
        return scan_id


def spec_comment(comment, doc=None, writer=None):
    """
    make it easy to add spec-style comments in a custom plan

    These comments *only* go into the SPEC data file.

    PARAMETERS

    comment *string* :
        (optional)
        Comment text to be written.  SPEC expects it to be only one line!

    doc *string* :
        (optional)
        Bluesky RunEngine document type.
        One of: ``start descriptor event resource datum stop``
        (default: ``event``)

    writer *obj* :
        (optional)
        Instance of ``SpecWriterCallback()``,
        typically: ``specwriter = SpecWriterCallback()``
    """
    global specwriter  # such as: specwriter = SpecWriterCallback()
    writer = writer or specwriter  # FIXME: get from namespace
    if doc is None:
        if writer.scanning:
            doc = "event"
        else:
            doc = "start"
    for line in comment.splitlines():
        writer._cmt(doc, line)

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



from ophyd import EpicsSignal
import pytest

from ..sseq import SseqRecord
from ..sseq import sseqRecordStep
from ..sseq import UserStringSequenceDevice
from ...tests import common_attribute_quantities_test
from ...tests import IOC
from ...tests import short_delay_for_EPICS_IOC_database_processing


@pytest.mark.parametrize(
    "device, pv, connect, attr, expected",
    [
        [SseqRecord, f"{IOC}userStringSeq10", False, "read_attrs", 31],
        [SseqRecord, f"{IOC}userStringSeq10", False, "configuration_attrs", 108],
        [SseqRecord, f"{IOC}userStringSeq10", True, "read()", 20],
        [SseqRecord, f"{IOC}userStringSeq10", True, "summary()", 274],

        [UserStringSequenceDevice, IOC, False, "read_attrs", 320],
        [UserStringSequenceDevice, IOC, False, "configuration_attrs", 1101],
        [UserStringSequenceDevice, IOC, True, "read()", 200],
        [UserStringSequenceDevice, IOC, True, "summary()", 2616],
    ]
)
def test_attribute_quantities(device, pv, connect, attr, expected):
    """Verify the quantities of the different attributes."""
    common_attribute_quantities_test(device, pv, connect, attr, expected)


def test_sseq_reset():
    user = UserStringSequenceDevice(IOC, name="user")
    user.wait_for_connection()
    user.enable.put("Enable")
    assert len(user.read()) == 200

    sseq = user.sseq10
    assert isinstance(sseq, SseqRecord)
    sseq.enable.put("E")  # Note: only "E"

    step = sseq.steps.step1
    assert isinstance(step, sseqRecordStep)

    step.reset()
    short_delay_for_EPICS_IOC_database_processing()

    assert step.input_pv.get() == ""

    uptime = EpicsSignal(f"{IOC}UPTIME", name="uptime")
    uptime.wait_for_connection()

    step.input_pv.put(uptime.pvname)
    assert step.string_value.get() == f"{0:.5f}"
    sseq.process_record.put(1)
    assert step.string_value.get() <= uptime.get()

    user.reset()
    short_delay_for_EPICS_IOC_database_processing()
    assert step.input_pv.get() == ""
    assert step.string_value.get() == f"{0:.5f}"



#!/usr/bin/env python

import setuptools
import setuptools_scm

setuptools.setup(
    # https://github.com/pypa/setuptools_scm/#setuppy-usage-deprecated
    version=setuptools_scm.get_version(),
    use_scm_version=True,
    setup_requires=["setuptools_scm"],

    # confirm the version from command line:
    # python -m setuptools_scm
    # python setup.py --version
)

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



#!/usr/bin/env python

"""
read SPEC config file and convert to ophyd setup commands

output of ophyd configuration to stdout

*new in apstools release 1.1.7*

**USAGE**

.. code-block:: bash

    user@host ~ $ ./spec2ophyd.py
    usage: spec2ophyd.py [-h] [-v] configFileName
    spec2ophyd.py: error: the following arguments are required: configFileName

    user@host ~ $ ./spec2ophyd.py -h
    usage: spec2ophyd.py [-h] [-v] configFileName

    read SPEC config file and convert to ophyd setup commands

    positional arguments:
    configFileName  SPEC config file name

    optional arguments:
    -h, --help      show this help message and exit
    -v, --version   print version number and exit

"""

from collections import OrderedDict
import re


KNOWN_DEVICES = "PSE_MAC_MOT VM_EPICS_M1 VM_EPICS_PV VM_EPICS_SC".split()


class Spec2ophydBase(object):

    str_keys = []
    mne = ""
    name = ""

    def obj_keys_to_list(self):
        items = []
        keys = list(self.str_keys)

        if (
            "mne" in keys
            and "name" in keys
            and hasattr(self, "mne")
            and hasattr(self, "name")
            and self.mne == self.name
        ):
            keys.pop(keys.index("name"))  # redundant, do not show

        for k in keys:
            if hasattr(self, k):
                v = self.__getattribute__(k)
                items.append(f"{k}='{v}'")
        return items

    def __str__(self):
        items = self.obj_keys_to_list()
        return f"{self.__class__.__name__}({', '.join(items)})"


class SpecDevice(Spec2ophydBase):
    """
    SPEC configuration of a device, e.g. multi-channel motor controller

    SPEC "devices" are components which counters or motors to controllers
    """

    def __init__(self, config_text):
        """parse the line from the SPEC config file"""
        self.raw = config_text
        # VM_EPICS_M1    = 9idcLAX:m58:c0: 8
        nm, args = config_text.split("=")
        self.name = nm.strip()
        prefix, num_channels = args.split()
        self.prefix = prefix
        self.config_line = None
        self.index = None
        self.num_channels = int(num_channels)
        self.ophyd_device = None
        self.str_keys = (
            "config_line name index, prefix num_channels".split()
        )


class ItemNameBase(Spec2ophydBase):

    ignore = False
    str_keys = "mne config_line name"

    def item_name_value(self, item):
        if hasattr(self, item):
            return f"{item}={self.__getattribute__(item)}"

    def ophyd_config(self):
        return f"{self.mne} = {self} # FIXME: not valid ophyd"


class SpecSignal(ItemNameBase):
    """
    SPEC configuration of a single EPICS PV
    """

    def __init__(self, mne, nm, pvname, config_text):
        """data provided by caller"""
        self.raw = config_text
        self.mne = mne
        self.name = nm
        self.pvname = pvname
        self.device = None
        self.signal_name = "EpicsSignal"

        lr = config_text.split(sep="=", maxsplit=1)

        def pop_word(line, int_result=False):
            line = line.strip()
            pos = line.find(" ")
            l, r = line[:pos].strip(), line[pos:].strip()
            if int_result:
                l = int(l)
            return l, r

        self.ctrl, r = pop_word(lr[1])
        self.unit, r = pop_word(r, True)
        self.chan, r = pop_word(r, True)
        self.scale, r = pop_word(r, True)
        self.flags = pop_word(r)[0]

        self.str_keys = "mne config_line name pvname signal_name".split()
        self.cntpar = {}

    def setDevice(self, devices):
        if self.ctrl.startswith("EPICS_PV"):
            device_list = devices.get("VM_EPICS_PV")
            if device_list is not None:
                self.device = device_list[self.unit]
                self.pvname = self.device.prefix
                self.ophyd_device = "EpicsSignal"
        elif self.ctrl.startswith("NONE"):
            self.ignore = True

    def ophyd_config(self):
        s = (
            f"{self.mne} = {self.signal_name}"
            f"('{self.pvname}', name='{self.mne}', labels=('detectors',))"
        )
        suffix = None
        if "misc_par_1" in self.cntpar:
            suffix = self.cntpar.pop("misc_par_1")
            pvname = f"{self.device.prefix}{suffix}"
            s = (
                f"{self.mne} = EpicsSignal"
                f"('{pvname}', name='{self.mne}', labels=('detectors',))"
            )
        if self.mne != self.name:
            s += f"  # {self.name}"
        if self.ignore:
            s = "# NONE: " + s
        if len(self.cntpar) > 0:
            terms = [f"{k}={v}" for k, v in self.cntpar.items()]
            s += f" # {', '.join(terms)}"
        return s


class SpecMotor(ItemNameBase):
    """
    SPEC configuration of a motor channel
    """

    def __init__(self, config_text):
        """parse the line from the SPEC config file"""
        self.raw = config_text
        # Motor    ctrl steps sign slew base backl accel nada  flags   mne  name
        # MOT002 = EPICS_M2:0/3   2000  1  2000  200   50  125    0 0x003       my  my
        # MOT022 = MAC_MOT:1/0/0   2000  1  2000  200   50  125    0 0x003   suvgap  SlitUpVGap
        lr = config_text.split(sep="=", maxsplit=1)
        self.config_line = int(lr[0].strip("MOT"))

        def pop_word(line, int_result=False):
            line = line.strip()
            pos = line.find(" ")
            l, r = line[:pos].strip(), line[pos:].strip()
            if int_result:
                l = int(l)
            return l, r

        self.ctrl, r = pop_word(lr[1])
        self.steps, r = pop_word(r, True)
        self.sign, r = pop_word(r, True)
        self.slew, r = pop_word(r, True)
        self.base, r = pop_word(r, True)
        self.backl, r = pop_word(r, True)
        self.accel, r = pop_word(r, True)
        self.nada, r = pop_word(r, True)
        self.flags, r = pop_word(r)
        self.mne, self.name = pop_word(r)
        self.device = None
        self.pvname = None
        self.motpar = {}
        self.mac_parms = None
        self.macro_prefix = None
        self.str_keys = "mne config_line name macro_prefix mac_parms".split()

    def __str__(self):
        items = self.obj_keys_to_list()
        txt = self.item_name_value("pvname") or self.item_name_value(
            "ctrl"
        )
        if not txt.endswith("=None"):
            items.append(txt)
        return f"{self.__class__.__name__}({', '.join(items)})"

    def setDevice(self, devices):
        if self.ctrl.startswith("EPICS_M2"):
            device_list = devices.get("VM_EPICS_M1")
            if device_list is not None:
                uc_str = self.ctrl[len("EPICS_M2:"):]
                unit, chan = list(map(int, uc_str.split("/")))
                self.device = device_list[unit]
                self.pvname = "{}m{}".format(self.device.prefix, chan)
        elif self.ctrl.startswith("MAC_MOT"):
            device_list = devices.get("PSE_MAC_MOT")
            if device_list is not None:
                uc_str = self.ctrl[len("MAC_MOT:"):]
                unit, *self.mac_parms = list(map(int, uc_str.split("/")))
                self.device = device_list[unit]
                self.macro_prefix = self.device.prefix
                # TODO: what else?
        elif self.ctrl.startswith("NONE"):
            self.ignore = True

    def ophyd_config(self):
        s = (
            f"{self.mne} = EpicsMotor"
            f"('{self.pvname}', name='{self.mne}', labels=('motor',))"
        )
        suffix = None
        if "misc_par_1" in self.motpar:
            # suffix = self.motpar.pop("misc_par_1")
            suffix = self.motpar["misc_par_1"]
            pvname = f"{self.device.prefix}{suffix}"
            s = (
                f"{self.mne} = EpicsMotor"
                f"('{pvname}', name='{self.mne}', labels=('motor',))"
            )
        if self.pvname is None:
            if self.macro_prefix is not None:
                s = f"# Macro Motor: {self}"
            else:
                s = f"# {self.config_line}: {self.raw}"
        if self.mne != self.name:
            s += f"  # {self.name}"
        if len(self.motpar) > 0:
            terms = [f"{k}={v}" for k, v in self.motpar.items()]
            s += f" # {', '.join(terms)}"
        return s


class SpecCounter(ItemNameBase):
    """
    SPEC configuration of a counter channel

    In SPEC's config file, a single PV signal is described as a counter,
    attached to an EPICS_PV (as described by a VM_EPICS_PV device).
    """

    def __init__(self, config_text):
        """parse the line from the SPEC config file"""
        self.raw = config_text
        # # Counter   ctrl unit chan scale flags    mne  name
        # CNT000 = EPICS_SC  0  0 10000000 0x001      sec  seconds

        def pop_word(line, int_result=False):
            line = line.strip()
            pos = line.find(" ")
            l, r = line[:pos].strip(), line[pos:].strip()
            if int_result:
                l = int(l)
            return l, r

        l, r = pop_word(config_text)
        self.config_line = int(l.strip("CNT"))
        l, r = pop_word(r)  # ignore "="
        self.ctrl, r = pop_word(r)
        self.unit, r = pop_word(r, True)
        self.chan, r = pop_word(r, True)
        self.scale, r = pop_word(r, True)
        self.flags, r = pop_word(r)
        self.mne, self.name = pop_word(r)
        self.device = None
        self.pvname = None
        self.reported_pvs = []
        self.str_keys = "mne config_line name unit chan".split()
        self.cntpar = {}

    def __str__(self):
        items = self.obj_keys_to_list()
        txt = self.item_name_value("pvname")
        if txt is not None:
            items.append(txt)
        else:
            items.append(self.item_name_value("ctrl"))
        return f"{self.__class__.__name__}({', '.join(items)})"

    def setDevice(self, devices):
        if self.ctrl.startswith("EPICS_SC"):
            device_list = devices.get("VM_EPICS_SC")
            if device_list is not None:
                self.device = device_list[self.unit]
                # scalers are goofy,
                # SPEC uses 0-based numbering,
                # scaler uses 1-based
                self.pvname = "{}.S{}".format(
                    self.device.prefix, self.chan + 1
                )
        elif self.ctrl.startswith("EPICS_PV"):
            device_list = devices.get("VM_EPICS_PV")
            if device_list is not None:
                self.device = device_list[self.unit]
                self.pvname = self.device.prefix
                self.ophyd_device = "EpicsSignal"
        elif self.ctrl.startswith("NONE"):
            self.ignore = True

    def ophyd_config(self):
        s = f"# counter: {self.mne} = {self}"
        suffix = None
        if "misc_par_1" in self.cntpar:
            suffix = self.cntpar.pop("misc_par_1")
            if self.device is None:
                self.ignore = True
            else:
                prefix = self.device.prefix
                pvname = f"{prefix}{suffix}"
                s = f"{self.mne} = EpicsSignal("
                s += (
                    f"'{pvname}', name='{self.mne}',"
                    f" labels=('detectors',)"
                )
                s += ")"
        if self.ignore:
            s = f"# {self.config_line}: {self.raw}"
        if len(self.cntpar) > 0:
            terms = [f"{k}={v}" for k, v in self.cntpar.items()]
            s += f" # {', '.join(terms)}"
        return s


class SpecConfig(object):
    """
    SPEC configuration
    """

    def __init__(self, config_file):
        self.config_file = config_file
        self.devices = OrderedDict()
        self.scalers = []
        self.collection = []
        self.unhandled = []

    def find_pv_in_collection(self, pv):
        for obj in self.collection:
            if pv == obj.pvname:
                return obj

    def read_config(self, config_file=None):
        self.config_file = config_file or self.config_file
        motor = None
        counter = None
        with open(self.config_file, "r") as f:
            for line_number, line in enumerate(f.readlines()):
                line = line.strip()

                if line.startswith("#"):
                    continue

                word0 = line.split(sep="=", maxsplit=1)[0].strip()
                if word0 in KNOWN_DEVICES:
                    device = SpecDevice(line)
                    if device.name not in self.devices:
                        self.devices[device.name] = []
                    # 0-based numbering
                    device.config_line = line_number
                    device.index = len(self.devices[device.name])
                    self.devices[device.name].append(device)
                elif word0.startswith("CNTPAR:"):
                    item = self.collection[-1]  # most recent item
                    if isinstance(item, (SpecSignal, SpecCounter)):
                        k, v = line[len("CNTPAR:"):].split("=")
                        item.cntpar[k.strip()] = v.strip()
                elif word0.startswith("MOTPAR:"):
                    if motor is not None:
                        k, v = line[len("MOTPAR:"):].split("=")
                        motor.motpar[k.strip()] = v.strip()
                elif re.match(r"CNT\d*", line) is not None:
                    counter = SpecCounter(line)
                    if counter.ctrl == "EPICS_PV":
                        # fmt: off
                        signal = SpecSignal(
                            counter.mne, counter.name,
                            counter.pvname, line
                        )
                        # fmt: on
                        signal.setDevice(self.devices)
                        self.collection.append(signal)
                    else:
                        counter.setDevice(self.devices)
                        if counter.pvname is not None:
                            pvname = counter.pvname.split(".")[0]
                            if pvname not in self.scalers:
                                mne = pvname.lower().split(":")[-1]
                                scaler = SpecSignal(mne, mne, pvname, line)
                                scaler.signal_name = "ScalerCH"
                                self.scalers.append(pvname)
                                self.collection.append(scaler)

                        if counter.pvname is None:
                            scaler = None
                        else:
                            scaler = self.find_pv_in_collection(
                                counter.pvname.split(".")[0]
                            )
                        if scaler is not None:
                            # clock = scaler.channels.chan01.s
                            fmt = "%s = %s.channels.chan%02d.s"
                            reassignment = fmt % (
                                counter.mne,
                                scaler.name,
                                counter.chan + 1,
                            )
                            counter = reassignment
                        self.collection.append(counter)
                elif re.match(r"MOT\d*", line) is not None:
                    motor = SpecMotor(line)
                    motor.setDevice(self.devices)
                    self.collection.append(motor)
                else:
                    self.unhandled.append(line)


def create_ophyd_setup(spec_config):
    print('"""')
    print("ophyd commands from SPEC config file")
    print()
    print(f"file: {spec_config.config_file}")
    print()
    print("CAUTION: Review the object names below before using them!")
    print("    Some names may not be valid python identifiers")
    print("    or may be reserved (such as ``time`` or ``del``)")
    print("    or may be vulnerable to re-definition because")
    print("    they are short or common.")
    print('"""')
    print()
    print("from ophyd import EpicsMotor, EpicsSignal")
    print("from ophyd.scaler import ScalerCH")
    print()
    for device in spec_config.collection:
        if hasattr(device, "ophyd_config"):
            print(f"{device.ophyd_config()}")
            if (
                hasattr(device, "signal_name")
                and device.signal_name == "ScalerCH"
            ):
                print(f"{device.name}.select_channels(None)")
        else:
            print(device)


def get_apstools_version():
    """Get the version string from the parent package."""
    from .. import __version__
    return __version__


def get_options():
    """Handle command line arguments."""
    import argparse
    import os
    import sys

    version = get_apstools_version()

    parser = argparse.ArgumentParser(
        prog=os.path.split(sys.argv[0])[-1],
        description=__doc__.strip().splitlines()[0],
    )

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        help="print version number and exit",
        version=version,
    )

    parser.add_argument(
        "configFileName", type=str, help="SPEC config file name"
    )

    return parser.parse_args()


def main():
    args = get_options()
    spec_cfg = SpecConfig(args.configFileName)
    spec_cfg.read_config()
    create_ophyd_setup(spec_cfg)


if __name__ == "__main__":
    main()

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



from .. import file_log_handler
from .. import get_log_path
from .. import setup_IPython_console_logging
from .. import stream_log_handler
import logging
import pathlib
import pytest
import tempfile


@pytest.fixture(scope="function")
def tempdir():
    tempdir = tempfile.mkdtemp()
    return tempdir


def test_get_log_path(tempdir):
    pathlib.os.chdir(tempdir)
    path = pathlib.Path() / ".logs"
    assert not path.exists()

    log_path = get_log_path()
    assert log_path.exists()


def test_file_log_handler(tempdir):
    pathlib.os.chdir(tempdir)
    handler = file_log_handler("my_log_file")
    assert isinstance(handler, logging.Handler)

    log_path = get_log_path()
    assert log_path.exists()

    logfile = log_path / "my_log_file.log"
    assert logfile.exists()

    # TODO: more tests


def test_stream_log_handler():
    handler = stream_log_handler()
    assert isinstance(handler, logging.Handler)
    # TODO: more tests


def test_setup_IPython_console_logging(tempdir):
    pathlib.os.chdir(tempdir)
    setup_IPython_console_logging()

    log_path = get_log_path()
    assert log_path.exists()

    # unit tests do not run IPython so no logfile exists.
    logfile = log_path / "ipython_console.log"
    assert not logfile.exists()



"""
Alignment plans
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~lineup
   ~tune_axes
   ~TuneAxis
   ~TuneResults
"""

import datetime
import logging
import numpy as np
import pyRestTable

from bluesky import plans as bp
from bluesky import plan_stubs as bps
from bluesky import preprocessors as bpp
from bluesky.callbacks.fitting import PeakStats
from ophyd import Device, Component, Signal
from ophyd.scaler import ScalerCH, ScalerChannel

from .. import utils


logger = logging.getLogger(__name__)


def lineup(
    # fmt: off
    detectors, axis, minus, plus, npts,
    time_s=0.1, peak_factor=4, width_factor=0.8,
    feature="cen",
    rescan=True,
    bec=None,
    md=None,
    # fmt: on
):
    """
    Lineup and center a given axis, relative to current position.

    If first run identifies a peak, makes a second run to fine tune the result.

    .. index:: Bluesky Plan; lineup

    PARAMETERS

    detectors
        *[object]* or *object* :
        Instance(s) of ophyd.Signal (or subclass such as ophyd.scaler.ScalerChannel)
        dependent measurement to be maximized.  If a list, the first signal in the
        list will be used.

    axis
        movable *object* :
        instance of ophyd.Signal (or subclass such as EpicsMotor)
        independent axis to use for alignment

    minus
        *float* :
        first point of scan at this offset from starting position

    plus
        *float* :
        last point of scan at this offset from starting position

    npts
        *int* :
        number of data points in the scan

    time_s
        *float* :
        Count time per step (if detectors[0] is ScalerChannel object),
        other object types not yet supported.
        (default: 0.1)

    peak_factor
        *float* :
        peak maximum must be greater than ``peak_factor*minimum``
        (default: 4)

    width_factor
        *float* :
        fwhm must be less than ``width_factor*plot_range``
        (default: 0.8)

    feature
        *str* :
        One of the parameters returned by BestEffortCallback peak stats.
        (``cen``, ``com``, ``max``, ``min``)
        (default: ``cen``)

    rescan
        *bool* :
        If first scan indicates a peak, should a second scan refine the result?
        (default: ``True``)

    bec
        *object* :
        Instance of ``bluesky.callbacks.best_effort.BestEffortCallback``.
        (default: ``None``, meaning look for it from global namespace)

    EXAMPLE::

        RE(lineup(diode, foemirror.theta, -30, 30, 30, 1.0))
    """
    bec = bec or utils.ipython_shell_namespace().get("bec")
    if bec is None:
        raise ValueError("Cannot find BestEffortCallback() instance.")
    bec.enable_plots()

    if not isinstance(detectors, (tuple, list)):
        detectors = [detectors]
    det0 = detectors[0]
    logger.info("Using detector '%s' for lineup()", det0.name)

    # first, determine if det0 is part of a ScalerCH device
    scaler = None
    obj = det0.parent
    if isinstance(det0.parent, ScalerChannel):
        if hasattr(obj, "parent") and obj.parent is not None:
            obj = obj.parent
            if hasattr(obj, "parent") and isinstance(obj.parent, ScalerCH):
                scaler = obj.parent

    if scaler is not None:
        old_sigs = scaler.stage_sigs
        scaler.stage_sigs["preset_time"] = time_s
        scaler.select_channels([det0.name])

    if hasattr(axis, "position"):
        old_position = axis.position
    else:
        old_position = axis.get()

    def peak_analysis():
        aligned = False

        if det0.name not in bec.peaks[feature]:
            logger.error(
                "No statistical analysis of scan peak for feature '%s'!"
                "  (bec.peaks=%s, bec=%s)",
                feature, bec.peaks, bec
            )
            yield from bps.null()
        else:
            table = pyRestTable.Table()
            table.labels = ("key", "value")
            table.addRow(("axis", axis.name))
            table.addRow(("detector", det0.name))
            table.addRow(("starting position", old_position))
            for key in bec.peaks.ATTRS:
                target = bec.peaks[key][det0.name]
                if isinstance(target, tuple):
                    target = target[0]
                table.addRow((key, target))
            logger.info(f"alignment scan results:\n{table}")

            lo = bec.peaks["min"][det0.name][-1]  # [-1] means detector
            hi = bec.peaks["max"][det0.name][-1]  # [0] means axis
            fwhm = bec.peaks["fwhm"][det0.name]
            final = bec.peaks["cen"][det0.name]

            # local PeakStats object of our plot of det0 .v x
            ps = list(bec._peak_stats.values())[0][det0.name]
            # get the X data range as received by PeakStats
            x_range = abs(max(ps.x_data) - min(ps.x_data))

            if final is None:
                logger.error("centroid is None")
                final = old_position
            elif fwhm is None:
                logger.error("FWHM is None")
                final = old_position
            elif hi < peak_factor * lo:
                logger.error(
                    "no clear peak: %f < %f*%f", hi, peak_factor, lo
                )
                final = old_position
            elif fwhm > width_factor * x_range:
                logger.error(
                    "FWHM too large: %f > %f*%f",
                    fwhm,
                    width_factor,
                    x_range,
                )
                final = old_position
            else:
                aligned = True

            logger.info(
                "moving %s to %f  (aligned: %s)",
                axis.name,
                final,
                str(aligned),
            )
            yield from bps.mv(axis, final)

        # too sneaky?  We're modifying this structure locally
        bec.peaks.aligned = aligned
        bec.peaks.ATTRS = ("com", "cen", "max", "min", "fwhm")

    _md = dict(purpose="alignment")
    _md.update(md or {})
    yield from bp.rel_scan([det0], axis, minus, plus, npts, md=_md)
    yield from peak_analysis()

    if bec.peaks.aligned and rescan:
        # again, tweak axis to maximize
        _md["purpose"] = "alignment - fine"
        fwhm = bec.peaks["fwhm"][det0.name]
        yield from bp.rel_scan([det0], axis, -fwhm, fwhm, npts, md=_md)
        yield from peak_analysis()

    if scaler is not None:
        scaler.select_channels()
        scaler.stage_sigs = old_sigs


class TuneAxis(object):
    """
    tune an axis with a signal

    .. index:: Bluesky Device; TuneAxis

    This class provides a tuning object so that a Device or other entity
    may gain its own tuning process, keeping track of the particulars
    needed to tune this device again.  For example, one could add
    a tuner to a motor stage::

        motor = EpicsMotor("xxx:motor", "motor")
        motor.tuner = TuneAxis([det], motor)

    Then the ``motor`` could be tuned individually::

        RE(motor.tuner.tune(md={"activity": "tuning"}))

    or the :meth:`tune()` could be part of a plan with other steps.

    Example::

        tuner = TuneAxis([det], axis)
        live_table = LiveTable(["axis", "det"])
        RE(tuner.multi_pass_tune(width=2, num=9), live_table)
        RE(tuner.tune(width=0.05, num=9), live_table)

    Also see the jupyter notebook tited **Demonstrate TuneAxis()**
    in the :ref:`examples` section.

    .. autosummary::

       ~tune
       ~multi_pass_tune
       ~peak_detected

    SEE ALSO

    .. autosummary::

       ~tune_axes

    """

    _peak_choices_ = "cen com".split()

    def __init__(self, signals, axis, signal_name=None):
        self.signals = signals
        self.signal_name = signal_name or signals[0].name
        self.axis = axis
        self.tune_ok = False
        self.peaks = None
        self.peak_choice = self._peak_choices_[0]
        self.center = None
        self.stats = []

        # defaults
        self.width = 1
        self.num = 10
        self.step_factor = 4
        self.peak_factor = 4
        self.pass_max = 4
        self.snake = True

    def tune(self, width=None, num=None, peak_factor=None, md=None):
        """
        Bluesky plan to execute one pass through the current scan range

        .. index:: Bluesky Plan; TuneAxis.tune

        Scan self.axis centered about current position from
        ``-width/2`` to ``+width/2`` with ``num`` observations.
        If a peak was detected (default check is that max >= 4*min),
        then set ``self.tune_ok = True``.

        PARAMETERS

        width
            *float* :
            width of the tuning scan in the units of ``self.axis``
            Default value in ``self.width`` (initially 1)
        num
            *int* :
            number of steps
            Default value in ``self.num`` (initially 10)
        md
            *dict* :
            (optional)
            metadata
        """
        width = width or self.width
        num = num or self.num
        peak_factor = peak_factor or self.peak_factor

        if self.peak_choice not in self._peak_choices_:
            msg = "peak_choice must be one of {}, geave {}"
            msg = msg.format(self._peak_choices_, self.peak_choice)
            raise ValueError(msg)

        initial_position = self.axis.position
        # final_position = initial_position       # unless tuned
        start = initial_position - width / 2
        finish = initial_position + width / 2
        self.tune_ok = False

        tune_md = dict(
            width=width,
            initial_position=self.axis.position,
            time_iso8601=str(datetime.datetime.now()),
        )
        _md = {
            "tune_md": tune_md,
            "plan_name": self.__class__.__name__ + ".tune",
            "tune_parameters": dict(
                num=num,
                width=width,
                initial_position=self.axis.position,
                peak_choice=self.peak_choice,
                x_axis=self.axis.name,
                y_axis=self.signal_name,
            ),
            "motors": (self.axis.name,),
            "detectors": (self.signal_name,),
            "hints": dict(dimensions=[([self.axis.name], "primary")]),
        }
        _md.update(md or {})
        if "pass_max" not in _md:
            self.stats = []
        self.peaks = PeakStats(x=self.axis.name, y=self.signal_name)

        @bpp.subs_decorator(self.peaks)
        def _scan(md=None):
            yield from bps.open_run(md)

            position_list = np.linspace(start, finish, num)
            # fmt: off
            signal_list = list(self.signals)
            signal_list += [self.axis]
            # fmt: on
            for pos in position_list:
                yield from bps.mv(self.axis, pos)
                yield from bps.trigger_and_read(signal_list)

            final_position = initial_position
            if self.peak_detected(peak_factor=peak_factor):
                self.tune_ok = True
                if self.peak_choice == "cen":
                    final_position = self.peaks.cen
                elif self.peak_choice == "com":
                    final_position = self.peaks.com
                else:
                    final_position = None
                self.center = final_position

            # add stream with results
            # yield from add_results_stream()
            stream_name = "PeakStats"
            results = TuneResults(name=stream_name)

            results.tune_ok.put(self.tune_ok)
            results.center.put(self.center)
            results.final_position.put(final_position)
            results.initial_position.put(initial_position)
            results.set_stats(self.peaks)
            self.stats.append(results)

            if results.tune_ok.get():
                yield from bps.create(name=stream_name)
                try:
                    yield from bps.read(results)
                except ValueError as ex:
                    separator = " " * 8 + "-" * 12
                    print(separator)
                    print(f"Error saving stream {stream_name}:\n{ex}")
                    print(separator)
                yield from bps.save()

            yield from bps.mv(self.axis, final_position)
            yield from bps.close_run()

            results.report(stream_name)

        return (yield from _scan(md=_md))

    def multi_pass_tune(
        self,
        width=None,
        step_factor=None,
        num=None,
        pass_max=None,
        peak_factor=None,
        snake=None,
        md=None,
    ):
        """
        Bluesky plan for tuning this axis with this signal

        Execute multiple passes to refine the centroid determination.
        Each subsequent pass will reduce the width of scan by ``step_factor``.
        If ``snake=True`` then the scan direction will reverse with
        each subsequent pass.

        PARAMETERS

        width
            *float* :
            width of the tuning scan in the units of ``self.axis``
            Default value in ``self.width`` (initially 1)
        num
            *int* :
            number of steps
            Default value in ``self.num`` (initially 10)
        step_factor
            *float* :
            This reduces the width of the next tuning scan by the given factor.
            Default value in ``self.step_factor`` (initially 4)
        pass_max
            *int* :
            Maximum number of passes to be executed (avoids runaway
            scans when a centroid is not found).
            Default value in ``self.pass_max`` (initially 4)
        peak_factor
            *float* :
            peak maximum must be greater than ``peak_factor*minimum``
            (default: 4)
        snake
            *bool* :
            If ``True``, reverse scan direction on next pass.
            Default value in ``self.snake`` (initially True)
        md
            *dict* :
            (optional)
            metadata
        """
        width = width or self.width
        num = num or self.num
        step_factor = step_factor or self.step_factor
        snake = snake or self.snake
        pass_max = pass_max or self.pass_max
        peak_factor = peak_factor or self.peak_factor

        self.stats = []

        def _scan(width=1, step_factor=10, num=10, snake=True):
            for _pass_number in range(pass_max):
                logger.info(
                    "Multipass tune %d of %d", _pass_number + 1, pass_max
                )
                _md = {
                    "pass": _pass_number + 1,
                    "pass_max": pass_max,
                    "plan_name": self.__class__.__name__
                    + ".multi_pass_tune",
                }
                _md.update(md or {})

                yield from self.tune(
                    width=width, num=num, peak_factor=peak_factor, md=_md
                )

                if not self.tune_ok:
                    return
                if width > 0:
                    sign = 1
                else:
                    sign = -1
                width = sign * 2 * self.stats[-1].fwhm.get()
                if snake:
                    width *= -1

        return (
            yield from _scan(
                width=width, step_factor=step_factor, num=num, snake=snake
            )
        )

    def multi_pass_tune_summary(self):
        t = pyRestTable.Table()
        t.labels = "pass Ok? center width max.X max.Y".split()
        for i, stat in enumerate(self.stats):
            row = [
                i + 1,
            ]
            row.append(stat.tune_ok.get())
            row.append(stat.cen.get())
            row.append(stat.fwhm.get())
            x, y = stat.max.get()
            row += [x, y]
            t.addRow(row)
        return t

    def peak_detected(self, peak_factor=None):
        """
        returns True if a peak was detected, otherwise False

        The default algorithm identifies a peak when the maximum
        value is four times the minimum value.  Change this routine
        by subclassing :class:`TuneAxis` and override :meth:`peak_detected`.
        """
        peak_factor = peak_factor or self.peak_factor

        if self.peaks is None:
            return False
        self.peaks.compute()
        if self.peaks.max is None:
            return False

        ymax = self.peaks.max[-1]
        ymin = self.peaks.min[-1]
        ok = ymax >= peak_factor * ymin  # this works for USAXS@APS
        if not ok:
            logger.info("ymax < ymin * %f: is it a peak?", peak_factor)
        return ok


def tune_axes(axes):
    """
    Bluesky plan to tune a list of axes in sequence

    .. index:: Bluesky Plan; tune_axes

    EXAMPLE

    Sequentially, tune a list of preconfigured axes::

        RE(tune_axes([mr, m2r, ar, a2r])

    SEE ALSO

    .. autosummary::

       ~TuneAxis
    """
    for axis in axes:
        yield from axis.tune()


class TuneResults(Device):
    """
    Provides bps.read() as a Device

    .. index:: Bluesky Device; TuneResults
    """

    tune_ok = Component(Signal)
    initial_position = Component(Signal)
    final_position = Component(Signal)
    center = Component(Signal)
    # - - - - -
    x = Component(Signal)
    y = Component(Signal)
    cen = Component(Signal)
    com = Component(Signal)
    fwhm = Component(Signal)
    min = Component(Signal)
    max = Component(Signal)
    crossings = Component(Signal)
    peakstats_attrs = "x y cen com fwhm min max crossings".split()

    def report(self, title=None):
        keys = (
            self.peakstats_attrs
            + "tune_ok center initial_position final_position".split()
        )
        t = pyRestTable.Table()
        t.addLabel("key")
        t.addLabel("result")
        for key in keys:
            v = getattr(self, key).get()
            t.addRow((key, str(v)))
        if title is not None:
            print(title)
        print(t)

    def set_stats(self, peaks):
        for key in self.peakstats_attrs:
            v = getattr(peaks, key)
            if v is None:  # None cannot be mapped into json
                v = "null"  # but "null" can replace it
            if key in ("crossings", "min", "max"):
                v = np.array(v)
            getattr(self, key).put(v)

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



from .. import restorable_stage_sigs
from ophyd import Component, Device, Signal
import bluesky
import pytest


RE = bluesky.RunEngine()


class MyTestDevice(Device):
    signal = Component(Signal, value=0.0)


@pytest.fixture(scope="function")
def device():
    yield MyTestDevice(name="device")


def rewrites_stage_sigs(obj):
    obj.stage_sigs = dict(signal=-1)
    yield from bluesky.plan_stubs.null()


def preserves_original_stage_sigs(obj):
    @restorable_stage_sigs([obj])
    def _inner():
        yield from rewrites_stage_sigs(obj)

    yield from _inner()


def test_device(device):
    assert hasattr(device, "signal")
    assert isinstance(device.signal, Signal)
    assert hasattr(device, "stage_sigs")
    assert isinstance(device.stage_sigs, dict)
    assert len(device.stage_sigs) == 0
    assert "signal" not in device.stage_sigs


@pytest.mark.parametrize(
    "plan, expected",
    # fmt: off
    [
        [rewrites_stage_sigs, 1],
        [preserves_original_stage_sigs, 0],
    ]
    # fmt: on
)
def test_plans(plan, expected, device):
    assert len(device.stage_sigs) == 0

    RE(plan(device))
    assert len(device.stage_sigs) == expected



"""
APS Machine Parameters
+++++++++++++++++++++++++++++++++++++++

**APS machine parameters**

.. autosummary::

   ~ApsMachineParametersDevice
   ~ApsOperatorMessagesDevice
"""

from ophyd import Component
from ophyd import Device
from ophyd import EpicsSignalRO
from .aps_cycle import ApsCycleDM


class ApsOperatorMessagesDevice(Device):
    """
    General messages from the APS main control room.

    .. index:: Ophyd Device; ApsOperatorMessagesDevice
    """

    operators = Component(EpicsSignalRO, "OPS:message1", string=True)
    floor_coordinator = Component(EpicsSignalRO, "OPS:message2", string=True)
    fill_pattern = Component(EpicsSignalRO, "OPS:message3", string=True)
    last_problem_message = Component(EpicsSignalRO, "OPS:message4", string=True)
    last_trip_message = Component(EpicsSignalRO, "OPS:message5", string=True)
    # messages 6-8: meaning?
    message6 = Component(EpicsSignalRO, "OPS:message6", string=True)
    message7 = Component(EpicsSignalRO, "OPS:message7", string=True)
    message8 = Component(EpicsSignalRO, "OPS:message8", string=True)


class ApsMachineParametersDevice(Device):
    """
    Common operational parameters of the APS of general interest.

    .. index:: Ophyd Device; ApsMachineParametersDevice

    EXAMPLE::

        import apstools.devices as APS_devices
        APS = APS_devices.ApsMachineParametersDevice(name="APS")
        aps_current = APS.current

        # make sure these values are logged at start and stop of every scan
        sd.baseline.append(APS)
        # record storage ring current as secondary stream during scans
        # name: aps_current_monitor
        # db[-1].table("aps_current_monitor")
        sd.monitors.append(aps_current)

    The `sd.baseline` and `sd.monitors` usage relies on this global setup::

        from bluesky import SupplementalData
        sd = SupplementalData()
        RE.preprocessors.append(sd)

    .. autosummary::

        ~inUserOperations

    """

    current = Component(EpicsSignalRO, "S:SRcurrentAI")
    lifetime = Component(EpicsSignalRO, "S:SRlifeTimeHrsCC")
    aps_cycle = Component(ApsCycleDM)
    machine_status = Component(EpicsSignalRO, "S:DesiredMode", string=True)
    # In [3]: APS.machine_status.enum_strs
    # Out[3]:
    # ('State Unknown',
    #  'USER OPERATIONS',
    #  'Bm Ln Studies',
    #  'INJ Studies',
    #  'ASD Studies',
    #  'NO BEAM',
    #  'MAINTENANCE')
    operating_mode = Component(EpicsSignalRO, "S:ActualMode", string=True)
    # In [4]: APS.operating_mode.enum_strs
    # Out[4]:
    # ('State Unknown',
    # 'NO BEAM',
    # 'Injecting',
    # 'Stored Beam',
    # 'Delivered Beam',
    # 'MAINTENANCE')
    shutter_permit = Component(EpicsSignalRO, "ACIS:ShutterPermit", string=True)
    fill_number = Component(EpicsSignalRO, "S:FillNumber")
    orbit_correction = Component(EpicsSignalRO, "S:OrbitCorrection:CC")
    global_feedback = Component(EpicsSignalRO, "SRFB:GBL:LoopStatusBI", string=True)
    global_feedback_h = Component(EpicsSignalRO, "SRFB:GBL:HLoopStatusBI", string=True)
    global_feedback_v = Component(EpicsSignalRO, "SRFB:GBL:VLoopStatusBI", string=True)
    operator_messages = Component(ApsOperatorMessagesDevice)

    @property
    def inUserOperations(self):
        """
        determine if APS is in User Operations mode (boolean)

        Use this property to configure ophyd Devices for direct or simulated hardware.
        See issue #49 (https://github.com/BCDA-APS/apstools/issues/49) for details.

        EXAMPLE::

            APS = apstools.devices.ApsMachineParametersDevice(name="APS")

            if APS.inUserOperations:
                suspend_APS_current = bluesky.suspenders.SuspendFloor(APS.current, 2, resume_thresh=10)
                RE.install_suspender(suspend_APS_current)
            else:
                # use pseudo shutter controls and no current suspenders
                pass

        """
        # fmt: off
        return self.machine_status.get() in (
            1, "USER OPERATIONS",
            2, "Bm Ln Studies",
        )
        # fmt: on

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Generalized ophyd Device base class for preamplifiers.

.. autosummary::

   ~PreamplifierBaseDevice
"""

from ophyd import Component
from ophyd import Device
from ophyd import Signal


class PreamplifierBaseDevice(Device):
    """
    Generalized interface (base class) for preamplifiers.

    All subclasses of ``PreamplifierBaseDevice`` must define how to update the
    gain with the correct value from the amplifier. An example is
    :class:`~apstools._devices.srs570_preamplifier.SRS570_PreAmplifier`.

    :see: https://github.com/BCDA-APS/apstools/issues/544
    """

    gain = Component(Signal, kind="normal", value=1)

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



from ..misc import connect_pvlist
import pytest


@pytest.mark.parametrize(
    "pvlist, success",
    [
        ["ad:cam1:Acquire gp:UPTIME".split(), True],
    ]
)
def test_connect_pvlist(pvlist, success):
    assert isinstance(pvlist, list)
    pvdict = connect_pvlist(pvlist)
    assert isinstance(pvdict, dict)



try:
    from setuptools_scm import get_version

    __version__ = get_version()
    del get_version
except (LookupError, ModuleNotFoundError):
    from importlib.metadata import version

    __version__ = version("apstools")

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Linkam temperature controllers
++++++++++++++++++++++++++++++

.. autosummary::

   ~Linkam_CI94_Device
   ~Linkam_T96_Device
"""

from ophyd import Component
from ophyd import Device
from ophyd import EpicsSignal
from ophyd import EpicsSignalRO
from ophyd import EpicsSignalWithRBV
from ophyd import Signal

from . import PVPositionerSoftDoneWithStop


class Linkam_CI94_Device(Device):
    """
    Linkam model CI94 temperature controller

    EXAMPLE::

        ci94 = Linkam_CI94_Device("IOC:ci94:", name="ci94")
    """

    controller_name = "Linkam CI94"

    temperature = Component(
        PVPositionerSoftDoneWithStop,
        "",
        readback_pv="temp",
        setpoint_pv="setLimit",
        tolerance=1.0,
        kind="hinted",
    )

    units = Component(Signal, value="C", kind="config")

    temperature_in = Component(EpicsSignalRO, "tempIn", kind="omitted")
    # DO NOT USE: temperature2_in = Component(EpicsSignalRO, "temp2In", kind="omitted")
    # DO NOT USE: temperature2 = Component(EpicsSignalRO, "temp2")

    dsc_in = Component(EpicsSignalRO, "dscIn", kind="omitted")
    end_after_profile = Component(EpicsSignal, "endAfterProfile", kind="omitted")
    end_on_stop = Component(EpicsSignal, "endOnStop", kind="omitted")
    error_byte = Component(EpicsSignalRO, "errorByte", kind="omitted")
    gen_stat = Component(EpicsSignalRO, "genStat", kind="omitted")
    hold_control = Component(EpicsSignal, "hold", kind="omitted")
    pump_mode = Component(EpicsSignal, "pumpMode", kind="omitted")
    pump_speed = Component(EpicsSignalRO, "pumpSpeed", kind="omitted")
    pump_speed_in = Component(EpicsSignalRO, "pumpSpeedIn", kind="omitted")
    rate = Component(EpicsSignal, "setRate", kind="omitted")  # RPM

    speed = Component(EpicsSignal, "setSpeed", kind="config")
    # deg/min, speed 0 = automatic control

    start_control = Component(EpicsSignal, "start", kind="omitted")
    status = Component(EpicsSignalRO, "status", kind="omitted")
    status_in = Component(EpicsSignalRO, "statusIn", kind="omitted")
    stop_control = Component(EpicsSignal, "stop", kind="omitted")

    # clear_buffer = Component(EpicsSignal, "clearBuffer", kind="omitted")  # bo
    # scan_dis = Component(EpicsSignal, "scanDis", kind="omitted")          # bo
    # test = Component(EpicsSignal, "test", kind="omitted")                 # longout
    # d_cmd = Component(EpicsSignalRO, "DCmd", kind="omitted")              # ai
    # t_cmd = Component(EpicsSignalRO, "TCmd", kind="omitted")              # ai
    # dsc = Component(EpicsSignalRO, "dsc", kind="omitted")                 # calc

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # temperature component is the main value
        self.temperature.name = self.name


class T96Temperature(PVPositionerSoftDoneWithStop):

    actuate = Component(EpicsSignalWithRBV, "heating", kind="config", string=True)
    actuate_value = "On"


class Linkam_T96_Device(Device):
    """
    Linkam model T96 temperature controller

    EXAMPLE::

        tc1 = Linkam_T96("IOC:tc1:", name="tc1")
    """

    controller_name = "Linkam T96"

    temperature = Component(
        T96Temperature,
        "",
        readback_pv="temperature_RBV",
        setpoint_pv="rampLimit",
        tolerance=1.0,
        kind="hinted",
    )

    ramprate = Component(EpicsSignalWithRBV, "rampRate", kind="config")
    units = Component(Signal, value="C", kind="config")

    controller_config = Component(EpicsSignalRO, "controllerConfig_RBV", kind="omitted")
    controller_error = Component(EpicsSignalRO, "controllerError_RBV", kind="omitted")
    controller_status = Component(EpicsSignalRO, "controllerStatus_RBV", kind="omitted")
    heater_power = Component(EpicsSignalRO, "heaterPower_RBV", kind="omitted")
    lnp_mode = Component(EpicsSignalWithRBV, "lnpMode", kind="omitted")
    lnp_speed = Component(EpicsSignalWithRBV, "lnpSpeed", kind="omitted")
    lnp_status = Component(EpicsSignalRO, "lnpStatus_RBV", kind="omitted")
    pressure = Component(EpicsSignalRO, "pressure_RBV", kind="omitted")
    ramp_at_limit = Component(EpicsSignalRO, "rampAtLimit_RBV", kind="omitted")
    stage_config = Component(EpicsSignalRO, "stageConfig_RBV", kind="omitted")
    status_error = Component(EpicsSignalRO, "statusError_RBV", kind="omitted")
    vacuum = Component(EpicsSignal, "vacuum", kind="omitted")
    vacuum_at_limit = Component(EpicsSignalRO, "vacuumAtLimit_RBV", kind="omitted")
    vacuum_limit_readback = Component(EpicsSignalWithRBV, "vacuumLimit", kind="omitted")
    vacuum_status = Component(EpicsSignalRO, "vacuumStatus_RBV", kind="omitted")  # calc

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # temperature component is the main value
        self.temperature.name = self.name

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



from .callback_base import FileWriterCallbackBase
from .doc_collector import DocumentCollectorCallback
from .doc_collector import document_contents_callback
from .nexus_writer import NEXUS_FILE_EXTENSION
from .nexus_writer import NEXUS_RELEASE
from .nexus_writer import NXWriter
from .nexus_writer import NXWriterAPS
from .spec_file_writer import SpecWriterCallback
from .spec_file_writer import spec_comment
from .spec_file_writer import SCAN_ID_RESET_VALUE
from .spec_file_writer import SPEC_TIME_FORMAT

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



import os
import ophyd
import ophyd.areadetector.filestore_mixins
import ophyd.areadetector.plugins
import pytest

from ... import utils

AD_IOC_PREFIX = "ad:"
GP_IOC_PREFIX = "gp:"
AD_IOC_FILES_ROOT = "/"
BLUESKY_FILES_ROOT = "/tmp/docker_ioc/ioc{AD_IOC_PREFIX.rstrip(':')}"
IOC_IMAGE_DIR = "/tmp/images/"
AD_IOC_PATH = os.path.join(AD_IOC_FILES_ROOT, IOC_IMAGE_DIR.lstrip("/"))
BLUESKY_PATH = os.path.join(BLUESKY_FILES_ROOT, IOC_IMAGE_DIR.lstrip("/"))


class MyCam(ophyd.SimDetectorCam):
    pool_max_buffers = None


class MyHDF5Plugin(
    ophyd.areadetector.filestore_mixins.FileStoreHDF5IterativeWrite,
    ophyd.areadetector.plugins.HDF5Plugin_V34
    # ophyd.HDF5Plugin
):
    pass


class MyFixedImagePlugin(ophyd.ImagePlugin):
    pool_max_buffers = None


class MyDetector(ophyd.SimDetector):
    cam = ophyd.ADComponent(MyCam, "cam1:")
    hdf1 = ophyd.ADComponent(
        MyHDF5Plugin,
        "HDF1:",
        write_path_template=AD_IOC_PATH,
        read_path_template=BLUESKY_PATH,
    )
    image = ophyd.ADComponent(MyFixedImagePlugin, "image1:")


gpm1 = ophyd.EpicsMotor(f"{GP_IOC_PREFIX}m1", name="gpm1")
simdet = MyDetector(AD_IOC_PREFIX, name="simdet")
gpm1.wait_for_connection()
simdet.wait_for_connection()


@pytest.mark.parametrize(
    "finder, reference, expected",
    [
        [
            utils.findbypv,
            f"{AD_IOC_PREFIX}image1:ArrayData",
            dict(read=["simdet.image.array_data"], write=["simdet.image.array_data"]),
        ],
        [
            utils.findbypv,
            f"{AD_IOC_PREFIX}HDF1:FilePath_RBV",
            dict(read=["simdet.hdf1.file_path"], write=[]),
        ],
        [
            utils.findbypv,
            f"{AD_IOC_PREFIX}cam1:Acquire",
            dict(read=[], write=["simdet.cam.acquire"]),
        ],
        [
            utils.findbypv,
            f"{GP_IOC_PREFIX}m1.RBV",
            dict(read=["gpm1.user_readback"], write=[]),
        ],
        [utils.findbyname, "m1_user_setpoint", None],
        [utils.findbyname, simdet.hdf1.array_size.name, "simdet.hdf1.array_size"],
        [utils.findbyname, "simdet_hdf1_array_size", "simdet.hdf1.array_size"],
        [utils.findbyname, "simdet_image_dim1_sa", "simdet.image.array_size.dim1_sa"],
        [
            utils.findbyname,
            "simdet_cam_peak_width_peak_width_y",
            "simdet.cam.peak_width.peak_width_y",
        ],
        [utils.findbyname, gpm1.user_setpoint.name, "m1.user_setpoint"],
        [utils.findbyname, simdet.cam.acquire.name, "simdet.cam.acquire"],
    ],
)
def test_main(finder, reference, expected):
    answer = finder(reference, ns=dict(m1=gpm1, simdet=simdet))
    assert answer == expected



"""
Axis Tuner
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~AxisTunerException
   ~AxisTunerMixin
"""

import datetime
import logging
from bluesky import plan_stubs as bps
from collections import OrderedDict
from .mixin_base import DeviceMixinBase

logger = logging.getLogger(__name__)


class AxisTunerException(ValueError):
    """
    Exception during execution of `AxisTunerBase` subclass

    .. index:: Ophyd Exception; AxisTunerException
    """


class AxisTunerMixin(DeviceMixinBase):
    """
    Mixin class to provide tuning capabilities for an axis

    .. index:: Ophyd Device Mixin; AxisTunerMixin

    See the `TuneAxis()` example in this jupyter notebook:
    https://github.com/BCDA-APS/apstools/blob/master/docs/source/resources/demo_tuneaxis.ipynb

    HOOK METHODS

    There are two hook methods (`pre_tune_method()`, and `post_tune_method()`)
    for callers to add additional plan parts, such as opening or closing shutters,
    setting detector parameters, or other actions.

    Each hook method must accept a single argument:
    an axis object such as `EpicsMotor` or `SynAxis`,
    such as::

        def my_pre_tune_hook(axis):
            yield from bps.mv(shutter, "open")
        def my_post_tune_hook(axis):
            yield from bps.mv(shutter, "close")

        class TunableSynAxis(AxisTunerMixin, SynAxis): pass

        myaxis = TunableSynAxis(name="myaxis")
        mydet = SynGauss('mydet', myaxis, 'myaxis', center=0.21, Imax=0.98e5, sigma=0.127)
        myaxis.tuner = TuneAxis([mydet], myaxis)
        myaxis.pre_tune_method = my_pre_tune_hook
        myaxis.post_tune_method = my_post_tune_hook

        def tune_myaxis():
            yield from myaxis.tune(md={"plan_name": "tune_myaxis"})

        RE(tune_myaxis())
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tuner = None  # such as: apstools.plans.TuneAxis

        # Hook functions for callers to add additional plan parts
        # Each must accept one argument: axis object such as `EpicsMotor` or `SynAxis`
        self.pre_tune_method = self._default_pre_tune_method
        self.post_tune_method = self._default_post_tune_method

    def _default_pre_tune_method(self):
        """called before `tune()`"""
        logger.info(
            "{} position before tuning: {}".format(
                self.name, self.position
            )
        )
        yield from bps.null()

    def _default_post_tune_method(self):
        """called after `tune()`"""
        logger.info(
            "{} position after tuning: {}".format(self.name, self.position)
        )
        yield from bps.null()

    def tune(self, md=None, **kwargs):
        if self.tuner is None:
            msg = "Must define an axis tuner, none specified."
            msg += "  Consider using apstools.plans.TuneAxis()"
            raise AxisTunerException(msg)

        if self.tuner.axis is None:
            msg = "Must define an axis, none specified."
            raise AxisTunerException(msg)

        if md is None:
            md = OrderedDict()
        md["purpose"] = "tuner"
        md["datetime"] = datetime.datetime.now().isoformat(sep=" ")

        if self.tuner is not None:
            if self.pre_tune_method is not None:
                yield from self.pre_tune_method()

            yield from self.tuner.tune(md=md, **kwargs)

            if self.post_tune_method is not None:
                yield from self.post_tune_method()

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Eurotherm 2216e Temperature Controller
++++++++++++++++++++++++++++++++++++++

The 2216e is a temperature controller from Eurotherm.

.. autosummary::

   ~Eurotherm2216e

According to their website, the [Eurotherm 2216e Temperature
Controller](https://www.eurothermcontrollers.com/eurotherm-2216e-series-controller-now-obsolete/)
is obsolete.  Please see replacement
[EPC3016](https://www.eurothermcontrollers.com/eurotherm-epc3016-1-16-din-process-and-temperature-controller/)
in our [EPC3000 Series](https://www.eurothermcontrollers.com/epc3000-series).

New in apstools 1.6.0.
"""

import logging

from ophyd import Component
from ophyd import EpicsSignal
from ophyd import EpicsSignalRO
from ophyd import Signal

from .positioner_soft_done import PVPositionerSoftDoneWithStop

logger = logging.getLogger(__name__)


class Eurotherm2216e(PVPositionerSoftDoneWithStop):
    """
    Eurotherm 2216e Temperature Controller
    """

    # temperature value in dC, needs conversion: readback = 0.1 * sensor
    sensor = Component(EpicsSignalRO, "SetPointSensor", kind="omitted")
    readback = Component(Signal, value=0, kind="hinted")
    setpoint = Component(
        EpicsSignal,
        "SetPointTemp",
        kind="hinted",
        write_pv="SetPointTempWrite",
        put_complete=True,
    )
    # temp1 & temp2 PVs have no useful values : ignore here
    # temp1 = Component(EpicsSignalRO, "Temp1", kind="hinted")
    # temp2 = Component(EpicsSignalRO, "Temp2", kind="hinted")
    power = Component(
        EpicsSignal,
        "SetPointPower",
        kind="config",
        write_pv="SetPointPowerWrite",
        put_complete=True,
    )

    mode = Component(
        EpicsSignal,
        "SetPointSensor",
        kind="config",
        write_pv="ModeWrite",
        string=True,
    )
    program_number = Component(EpicsSignalRO, "ProgramNumber", kind="config")
    # program_status = Component(
    #     EpicsSignalRO,
    #     "ProgramStatus",
    #     kind="config",
    #     string=True
    # )

    target = None  # remove from PVPositionerSoftDoneWithStop

    def cb_sensor(self, *args, **kwargs):
        "units: Convert dC from sensor to C"
        self.readback.put(0.1 * self.sensor.get())

    def __init__(self, prefix="", *, tolerance=1, **kwargs):
        super().__init__(
            prefix=prefix,
            readback_pv="ignoreRBV",
            setpoint_pv="ignore",
            tolerance=tolerance,
            update_target=False,
            **kwargs
        )
        self.sensor.subscribe(self.cb_sensor)

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Ophyd support for the EPICS epid record


Public Structures

.. autosummary::

    ~EpidRecord

:see: https://epics.anl.gov/bcda/synApps/std/epidRecord.html
"""

from ophyd import Component
from ophyd import EpicsSignal
from ophyd import EpicsSignalRO

from ._common import EpicsRecordDeviceCommonAll
from ._common import EpicsRecordFloatFields


class EpidRecord(EpicsRecordFloatFields, EpicsRecordDeviceCommonAll):
    """
    EPICS synApps epid record support in ophyd

    .. index:: Ophyd Device; synApps EpidRecord

    :see: https://epics.anl.gov/bcda/synApps/std/epidRecord.html
    """

    controlled_value_link = Component(EpicsSignal, ".INP", kind="config")
    controlled_value = Component(EpicsSignalRO, ".CVAL", kind="config")

    readback_trigger_link = Component(EpicsSignal, ".TRIG", kind="config")
    readback_trigger_link_value = Component(EpicsSignal, ".TVAL", kind="config")

    setpoint_location = Component(EpicsSignal, ".STPL", kind="config")
    setpoint_mode_select = Component(EpicsSignal, ".SMSL", kind="config")

    output_location = Component(EpicsSignal, ".OUTL", kind="config")
    feedback_on = Component(EpicsSignal, ".FBON", kind="config")

    proportional_gain = Component(EpicsSignal, ".KP", kind="config")
    integral_gain = Component(EpicsSignal, ".KI", kind="config")
    derivative_gain = Component(EpicsSignal, ".KD", kind="config")

    following_error = Component(EpicsSignalRO, ".ERR", kind="config")
    output_value = Component(EpicsSignalRO, ".OVAL", kind="config")
    final_value = Component(EpicsSignalRO, ".VAL", kind="normal")

    calculated_P = Component(EpicsSignalRO, ".P", kind="config")
    calculated_I = Component(EpicsSignal, ".I", kind="config")
    calculated_D = Component(EpicsSignalRO, ".D", kind="config")

    time_difference = Component(EpicsSignal, ".DT", kind="config")
    minimum_delta_time = Component(EpicsSignal, ".MDT", kind="config")

    # limits imposed by the record support:
    #     .LOPR <= .OVAL <= .HOPR
    #     .LOPR <= .I <= .HOPR
    high_limit = Component(EpicsSignal, ".DRVH", kind="config")
    low_limit = Component(EpicsSignal, ".DRVL", kind="config")

    @property
    def value(self):
        return self.output_value.get()

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Common support of slits

.. autosummary::

    ~SlitGeometry
"""

from collections import namedtuple


SlitGeometry = namedtuple("SlitGeometry", "width height x y")
"Slit size and center as a named tuple"

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Test the Excel support.
"""

import pathlib
import pytest

from ...plans import command_list
from ...utils import ExcelDatabaseFileGeneric


DATA_PATH = (pathlib.Path(command_list.__file__).parent / "tests")


@pytest.fixture(scope="function")
def xl_file():
    xl_file = (DATA_PATH / "demo3.xlsx")
    return xl_file


def test_testfile_exists(xl_file):
    assert DATA_PATH.exists()
    assert xl_file.exists()


def test_normal_read(xl_file):
    xl = ExcelDatabaseFileGeneric(xl_file)
    assert len(xl.db) == 9  # rows
    assert len(xl.db["0"]) == 7  # columns
    assert "Unnamed: 7" in xl.db["0"]
    assert xl.db["0"]["Unnamed: 7"] == 8.0


def test_ignore_extra_false(xl_file):
    xl = ExcelDatabaseFileGeneric(xl_file, ignore_extra=False)
    assert len(xl.db) == 16  # rows
    assert len(xl.db["0"]) == 9  # columns



"""
Tracking Signal for Device coordination
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~TrackingSignal
"""

from ophyd import Signal


class TrackingSignal(Signal):
    """
    Non-EPICS signal for use when coordinating Device actions.

    .. index:: Ophyd Signal; TrackingSignal

    Signal to decide if undulator will be tracked while changing the
    monochromator energy.
    """

    def check_value(self, value):
        """
        Check if the value is a boolean.

        RAISES

        ValueError
        """
        if not isinstance(value, bool):
            raise ValueError("tracking is boolean, it can only be True or False.")

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Flyers for motor and scaler scans
++++++++++++++++++++++++++++++++++++++++++++

.. rubric:: Public

.. autosummary::
   :recursive:

   ~SignalValueStack
   ~FlyerBase
   ~ActionsFlyerBase
   ~ScalerMotorFlyer

.. rubric:: Private

.. autosummary::
   :recursive:

   ~_SMFlyer_Step_1
   ~_SMFlyer_Step_2
   ~_SMFlyer_Step_3

New in release 1.6.9
"""

from ..utils import run_in_thread
from ophyd import Component
from ophyd import Device
from ophyd import DeviceStatus
from ophyd import Signal
from ophyd.status import Status
import pysumreg
import logging
import time

logger = logging.getLogger(__name__)


class SignalValueStack:
    """
    Gather a list of (signal, value) pairs to be restored later.

    Signals and values will restored in reverse order from how they were
    pushed to the stack.

    Not a bluesky plan. This is blocking code, intended for use in a separate
    thread from the RunEngine.

    Similar to the ``stage()`` / ``unstage()`` algorithm, an
    instance of ``SignalValueStack()`` records signals and their values
    to be reset once some operations have finished.

    Since any signal cannot be staged twice, or staged from within a running
    plan, we need some support that will save/restore values that are changed
    during a plan's execution.
    Use this for internal operations (such as in a multi-step fly plan)
    to allow original values to be recovered from Signals.

    .. rubric:: Usage

    ::

        from apstools.devices import SignalValueStack
        import ophyd
        signal = ophyd.Signal(name="signal", value=1)
        original = SignalValueStack()

        # ... in the threaded code
        original.remember(signal) # remember the original value
        signal.put(2)  # make some change(s)
        original.restore()  # put it back as we found it
    """

    def __init__(self) -> None:
        self.clear()

    # TODO: Use the 'list' interface?
    def push(self, signal, value):
        """Add a (signal, value) pair on the stack."""
        self._cache.append((signal, value))

    def clear(self):
        """Empty the stack."""
        self._cache = []

    def remember(self, signal):
        """Push the signal's value on the stack."""
        self.push(signal, signal.get())

    def restore(self):
        """Write the signals in the stack, in reverse order."""
        for pair in reversed(self._cache):
            pair[0].put(pair[-1])
        self._cache = []


class FlyerBase(Device):
    """
    Canonical ophyd Flyer object.  Acquires one empty data event.
    """

    STREAM_NAME = "primary"

    def kickoff(self):
        """Start the flyer."""
        status = DeviceStatus(self)
        status.set_finished()  # ALWAYS in kickoff()
        return status

    def complete(self):
        """Wait for the flyer to complete."""
        status = DeviceStatus(self)
        status.set_finished()  # once the flyer is done
        return status

    def describe_collect(self):
        """Describe the data from :meth:`~collect()`."""
        logging.debug("describe_collect()")
        schema = {}
        return {self.STREAM_NAME: schema}

    def collect(self):
        """Data event(s)."""
        logging.debug("collect()")
        event = dict(time=time.time(), data={}, timestamps={})
        yield event


class ActionsFlyerBase(FlyerBase):
    """
    Call :meth:`~actions_thread` method from :meth:`~kickoff`.

    Method :meth:`~actions_thread()` will describe all the fly scan steps.
    Method :meth:`~collect` is changed to wait for :meth:`~actions_thread`
    to complete before signalling that the flyer is complete and data is ready
    to :meth:`~collect`.

    In :meth:`~actions_thread()`, wait for all fly scan actions to complete,
    then mark the status object as finished before returning.
    """

    def __init__(self, *args, **kwargs):
        self.status_actions_thread = None
        self.status_complete = None
        self.status_kickoff = None
        super().__init__(*args, **kwargs)

    def kickoff(self):
        """Start the flyer."""
        logging.debug("kickoff()")

        self.status_actions_thread = Status()
        self.status_complete = Status()
        self.status_kickoff = Status()

        logging.debug("starting actions thread")
        self.actions_thread()

        self.status_kickoff.set_finished()  # ALWAYS in kickoff()
        return self.status_kickoff

    def complete(self):
        """Wait for the flyer to complete."""
        logging.debug("complete()")

        self.status_actions_thread.wait()
        self.status_complete.set_finished()
        logging.debug("completed actions thread")

        return self.status_complete

    def actions_thread(self):
        """
        Run the flyer in a thread.  Not a bluesky plan.

        Any acquired data should be saved internally and yielded from
        :meth:`~collect`.  (Remember to modify :meth:`~describe_collect` for
        any changes in the structure of data yielded by :meth:`~collect`!)
        """

        @run_in_thread
        def example_action():
            logging.debug("in actions thread")
            time.sleep(1)  # as a demonstration of a slow action
            self.status_actions_thread.set_finished()
            logging.debug("actions thread marked 'finished'")

        example_action()


class _SMFlyer_Step_1(ActionsFlyerBase):
    """
    Add a motor to ActionsFlyerBase().

    .. rubric:: Parameters

    motor
        *object* :
        Instance of :class:`ophyd.EpicsMotor`.
    start
        *float* or *int* :
        Fly scan will begin at this motor position.
    finish
        *float* or *int* :
        Fly scan will end at this motor position.
    fly_time
        *float* or *int* :
        Time (seconds, approximate) for fly scan to traverse from 'start' to
        'finish'.  The motor velocity will be set to meet this parameter.
        (Default: 1 s)
    fly_time_pad
        *float* or *int* :
        Extra time (seconds) to allow the fly scan to finish before a fly
        scan timeout is declared.
        (Default: 2 s)

    .. note:: This class is used internally to build, in steps,
        :class:`~ScalerMotorFlyer` from :class:`~ActionsFlyerBase`.
    """

    def __init__(self, motor, start, finish, *args, fly_time=1, fly_time_pad=2, **kwargs):
        fly_time = fly_time or 1
        fly_time_pad = fly_time_pad or 2
        if not hasattr(motor, "velocity"):
            raise TypeError(f"Unprepared to handle as motor: {motor=}")
        if fly_time <= 0:
            raise ValueError(f"Must be a POSITIVE number: {fly_time=}")
        if fly_time_pad <= 0:
            raise ValueError(f"Must be a POSITIVE number: {fly_time_pad=}")

        self._motor = motor
        self._pos_start = start
        self._pos_finish = finish
        self._fly_time = fly_time
        self._fly_time_pad = fly_time_pad

        self._original_values = SignalValueStack()

        self.status_fly = None
        self.status_taxi = None

        super().__init__(*args, **kwargs)

    @run_in_thread
    def actions_thread(self):
        """
        Prep and run the fly scan in a thread.

        Runs in background thread so it does not block RunEngine. This is not a
        generator function. It is not a bluesky plan, so blocking code is ok.
        """

        @run_in_thread
        def example_fly_scan():
            self.status_taxi = self._motor.move(self._pos_start, wait=False)
            self.status_taxi.wait()
            if self._motor.position != self._pos_start:  # TODO: within tolerance?
                raise RuntimeError(
                    "Not in requested taxi position:"
                    f" requested={self._pos_start}"
                    f" position={self._motor.position}"
                )

            velocity = abs(self._pos_finish - self._pos_start) / self._fly_time
            if velocity != self._motor.velocity.get():
                self._original_values.remember(self._motor.velocity)
                self._motor.velocity.put(velocity)

            self.status_fly = self._motor.move(self._pos_finish, wait=False)
            self.status_fly.wait(timeout=(self._fly_time + self._fly_time_pad))

            self._original_values.restore()
            self.status_actions_thread.set_finished()

        example_fly_scan()


class _SMFlyer_Step_2(_SMFlyer_Step_1):
    """
    Add modes (states) to _SMFlyer_Step_1() to recover from exceptions.

    Overrides :meth:`~actions_thread()` to recover from exceptions and restore
    previous values.  Since exceptions from threaded code are not reported
    through to the RunEgine (the thread simply stops for unhandled exceptions),
    the override handles such exceptions and then restores any signals to
    original values, as recorded via use of :class:`~SignalValueStack`. Here's
    the algorithm::

        try
            setup, taxi, then fly
        except Exception
            report the exception and save as attribute
        finally
            restore and return to idle

        mark status finished

    .. note:: This class is used internally to build, in steps,
        :class:`~ScalerMotorFlyer` from :class:`~ActionsFlyerBase`.
    """

    mode = Component(Signal, value="idle")
    ACTION_MODES = "idle setup taxi fly return".split()
    action_exception = None

    def actions_thread(self):
        """Run the flyer in a thread."""

        @run_in_thread
        def fly_scan_workflow():
            self.action_exception = None
            try:
                self._action_setup()
                self._action_taxi()
                self._action_fly()
            except Exception as exc:
                self.action_exception = exc
                logger.exception("Flyer Error")
                print(f"Flyer Error: {exc=}")
            finally:
                self._action_return()
                self.mode.put("idle")

            self.status_actions_thread.set_finished()

        fly_scan_workflow()

    def _action_setup(self):
        """Prepare for the taxi & fly."""
        self.mode.put("setup")
        self.status_taxi = None  # re-created by the taxi move
        self.status_fly = None  # re-created by the fly move
        self._original_values.clear()

    def _action_taxi(self):
        """Move motor to start position."""
        self.mode.put("taxi")
        self.status_taxi = self._motor.move(self._pos_start, wait=False)
        self.status_taxi.wait()
        if self._motor.position != self._pos_start:  # TODO: within tolerance?
            raise RuntimeError(
                "Not in requested taxi position:"
                f" requested={self._pos_start}"
                f" position={self._motor.position}"
            )

    def _action_fly(self):
        """Start the fly scan and wait for it to complete."""
        self.mode.put("fly")

        # set the fly scan velocity
        velocity = abs(self._pos_finish - self._pos_start) / self._fly_time
        if velocity != self._motor.velocity.get():
            self._original_values.remember(self._motor.velocity)
            self._motor.velocity.put(velocity)

        # get the motor moving
        self.status_fly = self._motor.move(self._pos_finish, wait=False)

        # wait for motor to be done moving
        allowed_motion_time = self._fly_time + self._fly_time_pad
        self.status_fly.wait(timeout=allowed_motion_time)

    def _action_return(self):
        """Restore original values."""
        self.mode.put("return")
        self._original_values.restore()


class _SMFlyer_Step_3(_SMFlyer_Step_2):
    """
    Add a scaler to _SMFlyer_Step_2() and trigger it for the fly motion.

    .. rubric:: Parameters

    scaler
        *object* :
        Instance of :class:`ophyd.scaler.ScalerCH`.
    scaler_time_pad
        *float* or *int* :
        Extra time (seconds) to leave the scaler counting before a fly scan
        timeout is declared.
        (Default: 2 s)

    .. note:: This class is used internally to build, in steps,
        :class:`~ScalerMotorFlyer` from :class:`~ActionsFlyerBase`.
    """

    def __init__(self, scaler, *args, scaler_time_pad=2, **kwargs):
        if not (
            # fmt: off
            hasattr(scaler, "count")
            and hasattr(scaler, "preset_time")
            and hasattr(scaler, "update_rate")
            # fmt: on
        ):
            raise TypeError(f"Unprepared to handle as scaler: {scaler=}")
        self._scaler = scaler
        self._scaler_time_pad = scaler_time_pad

        super().__init__(*args, **kwargs)

    def _action_fly(self):
        """Start the fly scan and wait for it to complete."""
        self.mode.put("fly")

        motion_time_allowance = self._fly_time + self._fly_time_pad
        count_time_allowance = self._fly_time + self._scaler_time_pad
        self._original_values.remember(self._scaler.count)
        self._original_values.remember(self._scaler.preset_time)
        self._scaler.preset_time.put(count_time_allowance)

        self._scaler.count.put("Count")  # start scaler counting
        time.sleep(1 / self._scaler.update_rate.get())

        # get the motor moving
        self.status_fly = self._motor.move(self._pos_finish, wait=False)

        # wait for motor to be done moving
        self.status_fly.wait(timeout=motion_time_allowance)
        self._scaler.count.put("Done")  # stop scaler counting


class ScalerMotorFlyer(_SMFlyer_Step_3):
    """
    Motor and scaler flyer, moving at constant velocity.

    Built from the :class:`~ActionsFlyerBase`, the
    :class:`~ScalerMotorFlyer` runs a fly scan described
    in :meth:`~actions_thread`.

    .. autosummary::

        ~_action_acquire_event
        ~_action_setup
        ~_action_fly
        ~_action_readings_to_collect_events
        ~describe_collect
        ~collect

    .. rubric:: Parameters

    period
        *float* or *int* :
        Time (seconds) between data collection events.
        (Default: 0.1 s)

    Extends :class:`SMFlyer__Step_3()`, adding periodic data acquisition.
    """

    def __init__(self, *args, period=None, **kwargs):
        period = period or 0.1
        if period <= 0:
            raise ValueError(
                f"Sampling period must be a POSITIVE number: {period=}"
            )

        self._period = period
        self._readings = []
        self.scaler_keys = None
        self.stats = None

        super().__init__(*args, **kwargs)

    def _action_acquire_event(self, **kwargs):
        """
        Record a single data acquisition event into the internal cache.

        Called by subscription to the scaler's elapsed counting time (CA monitor
        event).  Record the data as it exists when this is called.  Scaler
        timestamps are set from the current time. (Scaler timestamps update
        _only_ once counting has stopped.)
        """
        # read all the data at once
        ts = time.time()
        m_now = self._motor.read()
        s_now = self._scaler.read()

        for v in s_now.values():
            # Scaler timestamps update _only_ once counting has stopped.
            v["timestamp"] = ts  # substitute with current timestamp

        # Build the data acquisition event for collect() method.  Use _counting
        # and _done_moving keys to ID events for successive counter readings.
        event = dict(
            time=ts,
            _counting=self._scaler.count.get(),
            _done_moving=self._motor.motor_done_move.get(),
        )
        event.update(m_now)
        event.update(s_now)

        self._readings.append(event)  # add to internal cache

    def _action_setup(self):
        """Acquire new data at the scaler's update rate."""
        super()._action_setup()

        # try to set the sampling period
        expected_period = round(self._period, 5)
        self._original_values.remember(self._scaler.update_rate)
        self._scaler.update_rate.put(1 / self._period)
        received_period = round(
            1 / self._scaler.update_rate.get(use_monitor=False), 5
        )
        if received_period != expected_period:
            raise ValueError(
                "Could not get requested scaler sample period:"
                f" requested={self._period}"
                f" received={received_period}"
            )

        self.stats = None

    def _action_fly(self):
        """
        Start the fly scan and wait for it to complete.

        Data will be accumulated in response to CA monitor events from the
        scaler.
        """

        self.mode.put("fly")

        # set the fly scan velocity
        velocity = abs(self._pos_finish - self._pos_start) / self._fly_time
        if velocity != self._motor.velocity.get():
            self._original_values.remember(self._motor.velocity)
            self._motor.velocity.put(velocity)

        # set the scaler count time (allowance)
        self._original_values.remember(self._scaler.preset_time)
        count_time_allowance = self._fly_time + self._scaler_time_pad
        self._scaler.preset_time.put(count_time_allowance)

        # start acquiring, scaler update rate was set in _action_setup()
        self._scaler.time.subscribe(self._action_acquire_event)  # CA monitor

        # start scaler counting, THEN motor moving
        self._original_values.remember(self._scaler.count)
        self._scaler.count.put("Count")
        self.status_fly = self._motor.move(self._pos_finish, wait=False)

        # wait for motor to be done moving
        motion_time_allowance = self._fly_time + self._fly_time_pad
        self.status_fly.wait(timeout=motion_time_allowance)
        self._action_acquire_event()  # last event

        self._scaler.count.put("Done")  # stop scaler counting
        self._scaler.time.unsubscribe_all()  # stop acquiring

    def _action_readings_to_collect_events(self):
        """Return the readings as collect() events."""
        motor_keys = list(self._motor.read().keys())
        scaler_keys = list(self._scaler.read().keys())
        all_keys = motor_keys + scaler_keys
        self.stats = {k: pysumreg.SummationRegisters() for k in scaler_keys}

        previous = None
        for reading in self._readings:
            if (
                # fmt: off
                previous is not None
                and reading["_counting"]
                and not reading["_done_moving"]
                # fmt: on
            ):
                # reportable event
                ts = reading["time"]
                # fmt: off
                event = dict(
                    time=ts,
                    data={k: None for k in all_keys},
                    timestamps={k: None for k in all_keys},
                )
                # fmt: on

                for k in motor_keys:
                    event["data"][k] = reading[k]["value"]
                    event["timestamps"][k] = reading[k]["timestamp"]

                # use first motor reading is the one to use
                x_value = reading[motor_keys[0]]["value"]
                for k in scaler_keys:
                    # TODO: consider divide by time increment
                    delta = reading[k]["value"] - previous[k]["value"]
                    event["data"][k] = delta
                    event["timestamps"][k] = ts
                    self.stats[k].add(x_value, delta)

                yield event
            previous = reading.copy()

    def describe_collect(self):
        """Describe the data from collect()."""
        schema = {}
        schema.update(self._motor.describe())
        schema.update(self._scaler.describe())
        return {self.STREAM_NAME: schema}

    def collect(self):
        """Report the collected data."""
        from . import make_dict_device
        from bluesky import plan_stubs as bps

        yield from self._action_readings_to_collect_events()



"""
Diagnostic Support for Memory
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~rss_mem
"""

import os
import psutil


def rss_mem():
    """return memory used by this process"""
    return psutil.Process(os.getpid()).memory_info()

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



from ...tests import common_attribute_quantities_test
from ...tests import IOC
from ...tests import short_delay_for_EPICS_IOC_database_processing
from ..iocstats import IocStatsDevice
import pytest
import time


@pytest.mark.parametrize(
    "device, pv, connect, attr, expected",
    [
        [IocStatsDevice, IOC, False, "read_attrs", 19],
        [IocStatsDevice, IOC, False, "configuration_attrs", 8],
        [IocStatsDevice, IOC, True, "read()", 19],
        [IocStatsDevice, IOC, True, "summary()", 73],
    ]
)
def test_attribute_quantities(device, pv, connect, attr, expected):
    """Verify the quantities of the different attributes."""
    common_attribute_quantities_test(device, pv, connect, attr, expected)


@pytest.mark.parametrize(
    "attr, expected, comparison",
    [
        ["access", "Running", None],
        # TODO: ["application_directory", "iocxxx", "end"],
        ["engineer", "engineer", None],
        ["epics_version", "EPICS", "start"],
        ["kernel_version", "Linux", "start"],
        ["kernel_version", "x86_64", "end"],
        ["location", "location", None],
        ["records_count", 100, ">="],
    ]
)
def test_running(attr, expected, comparison):
    gp_info = IocStatsDevice(IOC, name="gp_info")
    gp_info.wait_for_connection()
    assert gp_info.connected
    short_delay_for_EPICS_IOC_database_processing()

    assert hasattr(gp_info, attr)
    value = getattr(gp_info, attr).get(use_monitor=False)
    time.sleep(.1)
    if comparison == "start":
        assert value.startswith(expected)
    elif comparison == "end":
        assert value.endswith(expected)
    elif comparison == "find":
        assert value.find(expected) >= 0
    elif comparison == ">=":
        assert value >= 0
    else:
        assert value == expected
